#!/usr/bin/env python

# exit codes
NOT_OK = 1
OK = 0

# modules for OPUS internal poller
import sys                  # access to command-line args
import opusutil             # utilites like PrintMsg
import re                   # regular expression handling
import traceback            # printing of stack traces 

# python modules for this application
import os
import getopt
import glob
import time
import math

# local modules for this application
import base_poller
import archdbquery
import rm_MissionFactory

# exceptions
class UsageError(Exception):
  pass

class UnknownRequestType(Exception):
  pass

class HaltEvent(Exception):
  pass

class DuplicateRequest(Exception):
  pass

class OSFCreationFailed(Exception):
  pass

class FailedEXTSearch(Exception):
  pass

class NoFilesFound(Exception):
  pass

class MissingRequiredResourceEntry(Exception):
  pass

class MissingExtensionValidation(Exception):
  pass

class UnsupportedExtValidationType(Exception):
  pass

class ExtensionValidationFailed(Exception):
  pass

class MismatchedRecordCount(Exception):
  pass

# indices into extension validation list
REQUIRED = 0
OPTIONAL = 1
FORBIDDEN = 2

# values for looking up PRODUCT and EXPOSURE long extensions in the db
FOR_PRODUCT = "Y"
FOR_EXPOSURE = "N"

# indices into ids query rows
IDS_DATASETNAME = 0
IDS_ARCHIVECLASS = 1
IDS_REQUESTID = 2
IDS_PATHNAME = 3

# retry modes
RETRY_NONE = 0
RETRY_GROUP_OSF = 1
RETRY_INSTALL_OSF = 2

# triggering modes
QUERY_LOG_MODE = 1
GROUP_MODE     = 0

class ReqManager(base_poller.InternalPoller):
  """
########################################################################
Class: ReqManager

Description:
------------
The class for interactive or pipeline ReqManager determination.

Members:
--------
ingest_path = root directory containing subdirs with files made available to NSA
pathname = name of path this task is running in (OSFs will be created in)
ingest_log_dir = location of dataset-specific log files
mission = name of mission being archived
pipe_id = 4-char string used in NSA request ids
file_name_formats = masks for reading file names for a particular 
                    mission/data_id
rm_control = contents of the control db rows for a given mission
product_long_ext = map of short extensions keyed on product long extensions
exposure_long_ext = map of short extensions keyed on exposure long extensions
MissionObj = object for mission-specific data and functions
Datasets = group of dataset objects
osf_create_status = status fields for creating Ingest pipeline OSFs
osf_update_status = status fields for updating Ingest pipeline OSFs
install_ids = list of install ids for the mission
cleanup_osf_stage = name of OSF stage triggered for cleanup in RETRY mode
cleanup_osf_status = status value for setting cleanup_osf_stage in RETRY mode
classesSkipExtValidation = comma-sep list of arch classes that SKIP extension
                           validation
arch_dbobj = database object for accessing ARCH_DB tables
arch_server = server name for archive database
arch_db = database name for archive database

Methods:
--------
usage() - prints an interactive usage message and raises an exception
initialize()
interactive() - handles command line args in interactive mode
pipelineInit() - handles resource parms in pipeline mode

History:
07/02/04  50703  MSwam   derive from base_poller, not internal_poller
06/19/06  56017  MSwam   add classesSkipExtValidation

########################################################################
  """
  def __init__(self):
    # initialize members with defaults
    self._ingest_path = None
    self._pathname    = None
    self._ingest_log_dir = None
    self._mission     = "KEPLER"
    self._pipe_id     = "DFLT"
    self._file_name_formats = {}
    self._rm_control  = {}
    self._product_long_ext = {}
    self._exposure_long_ext = {}
    self._MissionObj = None
    self._Datasets = []
    self._osf_create_status = "hw"
    self._osf_update_status = "hcw"
    self._install_ids = "KEP,KSI"
    self._cleanup_osf_stage = "CL"
    self._cleanup_osf_status = "c"
    self._classesSkipExtValidation = ""
    self._arch_dbobj       = None
    self._arch_server      = os.environ['ARCH_SERVER']
    self._arch_db          = os.environ['ARCH_DB']

  def initialize(self):
    #
    # call superclass constructor
    base_poller.InternalPoller.__init__(self)

  #--------------------------------------------------------------------------
  def usage(self):
    opusutil.PrintMsg('F','Interactive Usage: req_manager_main [-h] '
                          '-d datasetname -a data_id -p pathname '
                          '-g ingest_path -r resource_filespec '
                          '-o osf_create_status -u osf_update_status '
                          '-m mission')
    raise UsageError

  def commandLine(self):
    """
  #--------------------------------------------------------------------------
  #
  # Name: commandLine
  #
  # Description: Performs command line parsing for this task.
  #
  # Inputs: global sys.argv (command line arguments)
  #          
  # Outputs:
  #    dataset - name of dataset being processed
  #    data_id - data ID of dataset being processed
  #    resource_filespec - full path to resource file
  #
  # Notes:
  #        Will EXIT (status=1) if incorrect command line arguments
  #        are found.  This is a FATAL error.
  #
  # History:
  #   date    opr     who     reason
  # -------- -----  --------  --------------------------------------
  # 01/07/04 49551  MSwam     first version
  # 06/19/06 56017  MSwam     add classesSkipExtValidation, make a subroutine
  #
  #--------------------------------------------------------------------------
    """
    # Name: commandLine
    #
    # parse command line arguments
    try:
      # can't use "-i" (used by the OAPI)
      opts, args = getopt.getopt(sys.argv[1:], "hd:a:m:p:g:e:r:o:u:t:x:", 
                   ["help","datasetname=","data_id=","mission=","pathname=",
                    "ingest_path=","pipe_id=",
                    "resource_filespec=","osf_create_status=",
                    "osf_update_status=","install_ids=",
                    "classesSkipExtValidation="])
    except getopt.GetoptError:
      # print help information and raise exception
      self.usage()

    # initialize
    dataset = None
    data_id = None

    # parse options
    for flag, value in opts:
      if flag in ("-h", "--help"):
        self.usage()
      if flag in ("-d", "--datasetname"):
        dataset = value
      if flag in ("-a", "--data_id"):
        data_id = value
      if flag in ("-m", "--mission"):
        self._mission = value
      if flag in ("-p", "--pathname"):
        self._pathname = value
      if flag in ("-g", "--ingest_path"):
        self._ingest_path = value
      if flag in ("-e", "--pipe_id"):
        self._pipe_id = value
      if flag in ("-r", "--resource_filespec"):
        resource_filespec = value
      if flag in ("-o", "--osf_create_status"):
        self._osf_create_status = value
      if flag in ("-u", "--osf_update_status"):
        self._osf_update_status = value
      if flag in ("-t", "--install_ids"):
        self._install_ids = value
      if flag in ("-x", "--classesSkipExtValidation"):
        self._classesSkipExtValidation = value

    # verify required args
    if (dataset == None or data_id == None or self._ingest_path == None or
        resource_filespec == None):
      self.usage()
    #
    return (dataset, data_id, resource_filespec)


  def fillControlMaps(self):
    """
  #-------------------------------------------------------------------------
  # 
  # Name: fillControlMaps
  #
  # Description: Fills ingest control maps from DB tables.
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 01/07/04 49551  MSwam     first version
  # 07/02/04 50703  MSwam     use base_poller, add CLEANUP* items
  #
  #-------------------------------------------------------------------------
    """
    # def fillControlMaps(self):
    #
    # fill mission control map from the database
    self._rm_control = self._arch_dbobj.get_rm_control(self._mission)
    opusutil.PrintMsg("D","rm_control="+str(self._rm_control))

    # fill long extension maps from the database for the product and exposure
    self._product_long_ext = self._arch_dbobj.get_long_extensions(
                               self._mission, FOR_PRODUCT)
    opusutil.PrintMsg("D","product_long_ext="+str(self._product_long_ext))

    self._exposure_long_ext = self._arch_dbobj.get_long_extensions(
                               self._mission, FOR_EXPOSURE)
    opusutil.PrintMsg("D","exposure_long_ext="+str(self._exposure_long_ext))
    #
    return


  def interactive(self):
    """
  #--------------------------------------------------------------------------
  #
  # Name: interactive
  #
  # Description: Performs command line parsing and set-up for REQ_MANAGER
  #              processing in interactive mode.  Runs the REQ_MANAGER
  #              task for the command-line arguments given.
  #
  # Inputs: global sys.argv (command line arguments)
  #          
  # Notes:
  #        Will EXIT (status=1) if incorrect command line arguments
  #        are found.  This is a FATAL error.
  #
  # History:
  #   date    opr     who     reason
  # -------- -----  --------  --------------------------------------
  # 01/07/04 49551  MSwam     first version
  # 07/15/10 65561  MSwam     remove unused db retry parm
  #
  #--------------------------------------------------------------------------
    """
    # def interactive(self):
    #
    (dataset, data_id, resource_filespec) = self.commandLine()

    # connect to db
    self._arch_dbobj = archdbquery.Arch_db(self._arch_server,
                                         self._arch_db)

    # set the directory for file operations to the current directory
    self._ingest_log_dir = os.getcwd() + os.sep

    self.fillControlMaps()

    # read the resource file into a map
    resourceMap = opusutil.ResourceToMap(resource_filespec)

    # fill the file_name_format map from the resource entries
    self.fillFormatNames(resourceMap)
    opusutil.PrintMsg("D","file_name_formats="+str(self._file_name_formats))

    # create the mission object
    self._MissionObj = rm_MissionFactory.crank(self._mission)
    opusutil.PrintMsg("D","archive_class_map="+
                          str(self._MissionObj._archive_class_map))
    opusutil.PrintMsg("D","valid_extensions="+
                          str(self._MissionObj._valid_extensions))

    # run the task
    status = self.do_req_manager(dataset, data_id)

    # close the database connection
    self._arch_dbobj.close()

    return

  def fillSimpleResourceItems(self):
    """
  #-------------------------------------------------------------------------
  # 
  # Name: fillSimpleResourceItems
  #
  # Description: Reads resource file parameters for controlling the task.
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 01/07/04 49551  MSwam     first version
  # 07/02/04 50703  MSwam     use base_poller, add CLEANUP* items
  # 06/19/06 56017  MSwam     add classesSkipExtValidation
  #
  #-------------------------------------------------------------------------
    """
    # def fillSimpleResourceItems(self):
    #
    # obtain resource items in a dictionary and assign to class variables 
    try:
      resource_items = base_poller.InternalPoller.get_resource_items(self,
                            ["INGEST_PATH",
                             "INGEST_LOG_DIR",
                             "MISSION",
                             "PIPE_ID",
                             "OSF_CREATE",
                             "OSF_UPDATE",
                             "INSTALL_IDS",
                             "CLEANUP_OSF_STAGE",
                             "CLEANUP_OSF_STATUS",
                             "CLASSES_SKIP_EXT_VALIDATION",
                             "ARCH_DB",
                             "ARCH_SERVER",
                             "MISSION_DB",
                             "MISSION_DB"] )
    except:
      opusutil.PrintMsg("E","failed to get required resource entry")
      raise
    #
    opusutil.PrintMsg("I","Resource items read.")
    self._ingest_path = resource_items['INGEST_PATH']
    self._ingest_log_dir = resource_items['INGEST_LOG_DIR']
    self._mission = resource_items['MISSION']
    self._pipe_id = resource_items['PIPE_ID']
    self._osf_create_status = resource_items['OSF_CREATE']
    self._osf_update_status = resource_items['OSF_UPDATE']
    self._install_ids = resource_items['INSTALL_IDS']
    self._cleanup_osf_stage = resource_items['CLEANUP_OSF_STAGE']
    self._cleanup_osf_status = resource_items['CLEANUP_OSF_STATUS']
    self._classesSkipExtValidation = resource_items['CLASSES_SKIP_EXT_VALIDATION']
    self._arch_server = resource_items['ARCH_SERVER']
    self._arch_db = resource_items['ARCH_DB']
    #
    return(resource_items)


  def augmentResourceItems(self, resource_items):
    """
  #-------------------------------------------------------------------------
  # 
  # Name: augmentResourceItems
  #
  # Description: Adds resource items for each data_id's dataset name matching
  #              where they exist in the resource file.  Data_ids without
  #              corresponding resource items are acceptable, but incomplete 
  #              resource entries are not.
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 06/19/06 56017  MSwam     first version
  #
  #-------------------------------------------------------------------------
    """
    # def augmentResourceItems(self, resource_items):
    #
    # The control map contains the names of all file_name_format entries
    # that <might> be needed from the resource file.
    #
    # Try and get each set
    # but don't complain if one is not available (if that archive class
    # is not worked on by this invocation of the process, no problem will occur)
    #
    # DO complain if a set is incomplete (1st entry found, 2nd or 3rd not)
    #
    # Each row in the control map is keyed on data_id
    for k in self._rm_control.keys():
       # get the row in the control map for this data_id
       kmap = self._rm_control[k]
       format_name = kmap['irc_file_name_format']
       try:
         # file name format entries in the resource file come in a set of 3
         #   value,  value_REGEX, and value_REGEX_ID
         #
         # the 1st can be missing, but the 2nd and 3rd cannot
         resource_items[format_name] = self._papp.getResourceValue(format_name)
       except:
         opusutil.PrintMsg("D","No such resource entry: "+format_name)
         continue # skip out and don't expect the 2nd and 3rd entries
       try:
         reqd_name                   = format_name + "_REGEX"
         resource_items[reqd_name]   = self._papp.getResourceValue(reqd_name)
         reqd_name                   = format_name + "_REGEX_ID"
         resource_items[reqd_name]   = self._papp.getResourceValue(reqd_name)
       except:
         opusutil.PrintMsg("E","No such resource entry: "+reqd_name)
         raise MissingRequiredResourceEntry
    #
    return


  def pipelineInit(self):
    """
  #-------------------------------------------------------------------------
  # 
  # Name: pipelineInit
  #
  # Description: Initialization for pipeline mode, including reading
  #              resource file parameters for controlling the task.
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 01/07/04 49551  MSwam     first version
  # 07/02/04 50703  MSwam     use base_poller, add CLEANUP* items
  # 06/19/06 56017  MSwam     add classesSkipExtValidation
  # 07/15/10 65561  MSwam     remove unused db retry parm
  # 02/14/11 67507  Heller    Allow creation of File events to speed up INGEST
  #
  #-------------------------------------------------------------------------
    """
    # def pipelineInit(self):
    #
    resource_items = self.fillSimpleResourceItems()

    # get pathname this task is running in
    self._pathname = self._papp._get_path()

    opusutil.PrintMsg("I","Looking for USE_FILE_EVENT resource")
    try:
      tmp   = self._papp.getResourceValue('USE_FILE_EVENT')
      if tmp[0] == 'T' or  tmp[0] =="t":
        self._use_file_event="True"
        opusutil.PrintMsg("I","Using File Event to trigger OSF updates")
      else:
        self._use_file_event = None
        opusutil.PrintMsg("I","Using OSF BB to trigger OSF updates")   
    except:
      opusutil.PrintMsg("I","Defaulting to OSF BB updates ")
      self._use_file_event = None
    if (self._use_file_event !=None):
      try:
        self._kr_file_event =  self._papp.getResourceValue('KR_FILE_EVENT')
        opusutil.PrintMsg("I","File Events placed in "+self._kr_file_event)
      except:
        opusutil.PrintMsg("E","Missing resoure KR_FILE_EVENT")
        raise MissingRequiredResourceEntry


    # connect to the database
    self._arch_dbobj = archdbquery.Arch_db(self._arch_server,
                                         self._arch_db)
    self.fillControlMaps()

    self.augmentResourceItems(resource_items)

    # find the input file masks that should exist
    # in the resource file (mission-specific) and read them into a map
    #
    self.fillFormatNames(resource_items)
    opusutil.PrintMsg("D","file_name_formats="+str(self._file_name_formats))

    # create the mission object
    self._MissionObj = rm_MissionFactory.crank(self._mission)
    opusutil.PrintMsg("D","archive_class_map="+
                          str(self._MissionObj._archive_class_map))
    opusutil.PrintMsg("D","valid_extensions="+
                          str(self._MissionObj._valid_extensions))
    #
    return


  def fillFormatNames(self, theResources):
    """
  #-------------------------------------------------------------------------
  # 
  # Name: fillFormatNames
  #
  # Description: Fills the file_format_name map from the resource map.
  #
  # Inputs:
  #   theResources - a map of resource file name/value pairs
  #
  # Exceptions:
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 01/07/04 49551  MSwam     first version
  #
  #-------------------------------------------------------------------------
    """
    # def fillFormatNames(self, theResources):
    #
    # find the input file masks that should exist in the resource file 
    # (mission-specific, as indicated in the control information) 
    # and save them in a map, keyed on the format type
    #
    # Each row in the control map is keyed on data_id
    for k in self._rm_control.keys():
       # get the row in the control map for this data_id
       kmap = self._rm_control[k]
       format_name = kmap['irc_file_name_format']
       #
       # don't fill value again, if it was already done 
       #   (multiple data_ids can use the same format)
       if self._file_name_formats.has_key(format_name):
         continue
       #
       # Get the format mask, regular expression, and EXT field id from the 
       #   resource file for the file_name_format to be used for this data_id.
       # Save them in a map entry (each entry is a list of four), 
       #   keyed by the file_name_format.
       # The compiled regular expression is in the entry as well.
       # 
       # (the regex and regex_id values in the resource file are found
       #  by just appending fixed strings to the format name)
       #
       self._file_name_formats[format_name] = ([
                               theResources[format_name],
                               theResources[format_name+"_REGEX"],
                               theResources[format_name+"_REGEX_ID"],
                               re.compile(theResources[format_name+"_REGEX"]) ])
    #
    return


  def calcGenerationDate(self):
    """
  #-------------------------------------------------------------------------
  # 
  # Name: calcGenerationDate
  #
  # Description: Calculates the generation date (latest file creation date)
  #              for the dataset list.
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 01/07/04 49551  MSwam     first version
  #
  #-------------------------------------------------------------------------
    """
    # def calcGenerationDate(self):
    #
    gen_date = -1
    for d in self._Datasets:
      for f in d._Files:
        if f._file_creation_date > gen_date:
          gen_date = f._file_creation_date
    #
    return gen_date

  def nearestHundredths(self, time_now):
    """
  #-------------------------------------------------------------------------
  # 
  # Name: nearestHundredths
  #
  # Description: Retrieves the clock time hundredths of seconds for a
  #              given time.
  #
  # Inputs:
  #  time_now
  #
  # Returns:
  #  hundredths
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 06/22/06 56017  MSwam     first version
  #
  #-------------------------------------------------------------------------
    """
    # def nearestHundredths(self, time_now):
    #
    subseconds = time_now - int(time_now)
    hundredths = int(math.floor(subseconds * 100.0 + 0.5))
    if hundredths == 100:
      hundredths = 0
    opusutil.PrintMsg("D","time="+str(time_now)+" subsecs="+str(subseconds)+
                          " hundredths="+str(hundredths))
    #
    return hundredths
  

  def fillRequestID(self, time_now, hundredths, sequence_number):
    """
  #-------------------------------------------------------------------------
  # 
  # Name: fillRequestID
  #
  # Description: Builds a request ID from the ingest pipeline ID, current
  #              time to the nearest hundredth of a second, and a sequence
  #              number.
  #
  # Inputs:
  #   time_now
  #   hundredths
  #   sequence_number
  #
  # Returns:
  #   ingest_request_id
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 06/22/06 56017  MSwam     first version
  #
  #-------------------------------------------------------------------------
    """
    # def fillRequestID(self, time_now, hundredths, sequence_number):
    #
    #
    # format the time as needed and add hundredths (zero-filled to 2 places)
    timestamp = (time.strftime("%Y%m%d%H%M%S", time.localtime(time_now)) +
                 str(hundredths).zfill(2))
    #
    # add the seq number (zero-filled to 3 places)
    ingest_request_id = ("I"+self._pipe_id.upper()+timestamp+
                          str(sequence_number).zfill(3))
    #
    return ingest_request_id


  def calcInstallRequestID(self, time_now, sequence_number):
    """
  #-------------------------------------------------------------------------
  # 
  # Name: calcInstallRequestID
  #
  # Purpose: Calculates the value of the install request id, mostly
  #          using the current time.
  #
  # Description: An install request id value is calculated to meet this
  #              template:
  #                I + <PIPE_ID> + <TIMESTAMP> + <SEQNUM>
  #              where
  #                I = fixed character "I"
  #                PIPE_ID - pipeline id from task init
  #                TIMESTAMP - YYYYMMDDHHMMSSCC (CC = hundredths of sec)
  #                SEQNUM - zero-filled three digit value starting at 001
  #                         and incrementing for each dataset in a group
  #
  #              A query is made against the ingest_data_set_info table
  #              to ensure a request is not already using that ID. If a
  #              match is found, the hundredths are bumped and another
  #              search is made.  If it still fails an exception is raised.
  #
  # Inputs:
  #   time_now - current time in float number of seconds since 1970
  #   sequence_number - the seq number of this dataset group
  #
  # Returns:
  #   install_request_id
  #
  # Exceptions:
  #   NoUniqueRequestID
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 01/07/04 49551  MSwam     first version
  # 03/22/04 49551  MSwam     fix for hundredths = 100
  # 04/14/04 50823  MSwam     make time print debug-level
  #
  #-------------------------------------------------------------------------
    """
    # def calcInstallRequestID(self, time_now, sequence_number):
    #
    # extract the current subseconds to nearest 100th
    hundredths = self.nearestHundredths(time_now)

    ingest_request_id = self.fillRequestID(time_now, 
                                           hundredths, sequence_number)

    #
    # check for a duplicate
    result = self._arch_dbobj.count_ingest_data_set_info_reqid(
                                                         ingest_request_id)
    if result[0][0] > 0:
      #
      # DUPLICATE - bump the time and try again
      opusutil.PrintMsg("I","Duplicate entry already in MISSION_DB.ingest_data_set_info") 
      opusutil.PrintMsg("I","for "+ingest_request_id+". Trying another.")
      hundredths += 1
      if hundredths == 100:
        hundredths = 0

      ingest_request_id = self.fillRequestID(time_now,
                                             hundredths, sequence_number)
      #
      # check for a duplicate
      result = self._arch_dbobj.count_ingest_data_set_info_reqid(
                                                         ingest_request_id)
      if result[0][0] > 0:
        #
        # DUPLICATE
        opusutil.PrintMsg("I","Duplicate entry already in MISSION_DB.ingest_data_set_info") 
        raise NoUniqueRequestID
    #
    return (ingest_request_id)


  def checkForDuplicates(self, dataset, data_id, generation_date):
    """
  #-------------------------------------------------------------------------
  # 
  # Name: checkForDuplicates
  #
  # Purpose: Performs various tests to ensure that an archive request
  #          does not duplicate an existing request.
  #
  # Description: Checks the following databases and tables for existing
  #              matching entries:
  #                MISSION_DB.ingest_data_set_info
  #                ARCH_DB.archive_data_set_all
  #                ARCH_DB.archive_files
  #
  # Inputs:
  #   dataset
  #   data_id
  #   generation_date 
  #
  # Exceptions
  #   DuplicateRequest
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 01/07/04 49551  MSwam     first version
  #
  #-------------------------------------------------------------------------
    """
    # def checkForDuplicates(self, dataset, data_id, generation_date):
    #

    # first check the ingest_data_set_info table for a match
    result = self._arch_dbobj.count_ingest_data_set_info_ddgm(
                                                         data_id, dataset, 
                                                         generation_date,
                                                         self._mission)
    if result[0][0] > 0:
      #
      # DUPLICATE
      opusutil.PrintMsg("E","Duplicate entry already in MISSION_DB.ingest_data_set_info") 
      opusutil.PrintMsg("E","for "+dataset+","+data_id+","+generation_date)
      raise DuplicateRequest
    #
    # now check the archive_data_set_all and archive_files table 
    # for each dataset
    for d in self._Datasets:
      result = self._arch_dbobj.count_archive_data_set_all(
                                  self._mission, d._archive_class_name,
                                  d._dataset_name, generation_date)
      if result[0][0] > 0:
        #
        # DUPLICATE
        opusutil.PrintMsg("E","Duplicate entry already in ARCH_DB.archive_data_set_all") 
        opusutil.PrintMsg("E","for "+self._mission+","+d._archive_class_name+","+d._dataset_name+","+generation_date)
        raise DuplicateRequest
      #
      result = self._arch_dbobj.count_archive_files(
                                  self._mission, d._archive_class_name,
                                  d._dataset_name, generation_date)
      if result[0][0] > 0:
        #
        # DUPLICATE
        opusutil.PrintMsg("E","Duplicate entry already in ARCH_DB.archive_files") 
        opusutil.PrintMsg("E","for "+self._mission+","+d._archive_class_name+","+d._dataset_name+","+generation_date)
        raise DuplicateRequest
    #
    return


  def load_ingest_files(self, generation_date):
    """
  #-------------------------------------------------------------------------
  # 
  # Name: load_ingest_files
  #
  # Purpose: Loads the ingest_files table with rows for each
  #          file.
  #
  # Description:
  #
  # Inputs:
  #   generation_date
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 01/07/04 49551  MSwam     first version
  #
  #-------------------------------------------------------------------------
    """
    # def load_ingest_files(self, generation_date):
    #
    # insert a row for each file
    for d in self._Datasets:
      for f in d._Files:
        self._arch_dbobj.insert_ingest_files(d._archive_class_name,
                         d._dataset_name, generation_date, self._mission,
                         f._file_extension, f._file_name)
    #
    return


  def createOSFs(self, data_id):
    """
  #-------------------------------------------------------------------------
  # 
  # Name: createOSFs
  #
  # Purpose: Creates pipeline OSFs for each dataset entry.
  #
  # Inputs:
  #   data_id
  #
  # Exceptions:
  #   OSFCreationFailed
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 01/07/04 49551  MSwam     first version
  # 01/04/11 67158  heller    use update status and skip update to make faster
  # 02/14/11 67507  Heller    Allow creation of File events to speed up INGEST
  #-------------------------------------------------------------------------
    """
    # def createOSFs(self, data_id):
    #
    # create an OSF for each dataset
    for d in self._Datasets:
      osf_dataset = d._dads_request_id
      osf_data_id = self._rm_control[data_id]['irc_request_data_id']
      osf_status = self._osf_update_status
      #osf_status = self._osf_create_status
      # create OSF
      if (self._use_file_event == None):      
        try:
          opusutil.PrintMsg("I","creating OSF for dataset="+osf_dataset+
                               " dataid="+osf_data_id+
                               " path="+self._pathname)
          os.system('osf_create -p ' + self._pathname + ' -f ' + osf_dataset +
                   ' -t ' + osf_data_id + ' -n 000 -s '+
                   self._osf_update_status)
        except:
          opusutil.PrintMsg('E','OSF creation FAILED.')
          raise OSFCreationFailed
      else:
         opusutil.PrintMsg("I","Creating FILE event in "+self._kr_file_event) 
         filecreate = self._kr_file_event+"/"+osf_dataset+"."+osf_data_id+"_create"
         _file_c = open(filecreate, 'w')
         _file_c.write('OSF_UPDATE\n')
         _file_c.write("CREATE \n")
         _file_c.close()
         try:
           fileE = self._kr_file_event+"/"+osf_dataset+"."+osf_data_id+".ready"
           os.rename(filecreate, fileE)
         except OSError, (errno, errstr):
           opusutil.PrintMsg("E","Failed File Event Creation: "+fileE)
           opusutil.PrintMsg("E","rename failed: "+errstr)
           raise NoFilesFound
 
    #
    return


  def updateOSFs(self, data_id):
    """
  #-------------------------------------------------------------------------
  # 
  # Name: updateOSFs
  #
  # Purpose: Updates pipeline OSFs for each dataset entry.
  #
  # Inputs:
  #   data_id
  #
  # Exceptions:
  #   OSFUpdateFailed
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 01/07/04 49551  MSwam     first version
  #
  #-------------------------------------------------------------------------
    """
    # def updateOSFs(self, data_id):
    #
    # update an OSF for each dataset
    for d in self._Datasets:
      osf_dataset = d._dads_request_id
      osf_data_id = self._rm_control[data_id]['irc_request_data_id']
      osf_status = self._osf_update_status
      # update OSF
      try:
        opusutil.PrintMsg("I","updating OSF for dataset="+osf_dataset+
                               " dataid="+osf_data_id+
                               " path="+self._pathname)
        os.system('osf_update -p ' + self._pathname + ' -f ' + osf_dataset +
                   ' -t ' + osf_data_id + ' -n 000 -s '+
                   self._osf_update_status)
      except:
        opusutil.PrintMsg('E','OSF update FAILED.')
        raise OSFUpdateFailed
    #
    return


  def query_log(self, dataset):
    """
  #-------------------------------------------------------------------------
  # 
  # Name: query_log
  #
  # Purpose: Queries the ingest_data_set_info table for a particular
  #          dataset.
  #
  # Description: A query is made on ids_ins_request_id using the passed
  #              dataset name as key.  Field values are retrieved from
  #              the table and written to the log file.  If no record
  #              is found in the table for this key, an exception is raised.
  #
  # Inputs:
  #   dataset
  #
  # Exceptions:
  #   MissingInfoRecord
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 01/07/04 49551  MSwam     first version
  #
  #-------------------------------------------------------------------------
    """
    # def query_log(self, dataset):
    #
    # find the row in the table
    result = self._arch_dbobj.get_ingest_data_set_info_reqid(dataset)
    #
    # if not found, print error and raise exception
    if len(result) == 0:
      opusutil.PrintMsg("E","Failed to find ingest_data_set_info record for "+
                        dataset)
      raise MissingInfoRecord
    #
    # print the values, if found
    opusutil.PrintMsg("I","group_name     : "+ str(result[0][0]))
    opusutil.PrintMsg("I","group_data_id  : "+ str(result[0][1]))
    opusutil.PrintMsg("I","archive_class  : "+ str(result[0][2]))
    opusutil.PrintMsg("I","generation_date: "+ str(result[0][3]))
    opusutil.PrintMsg("I","mission        : "+ str(result[0][4]))
    opusutil.PrintMsg("I","receipt_date   : "+ str(result[0][5]))
    # 
    return


  def find_ext_validation(self, data_id):
    """
  #-------------------------------------------------------------------------
  # 
  # Name: find_ext_validation
  #
  # Description: Given a data_id, return a list of the
  #              extension validation information for that id
  #              The information is a list of 3 lists, one per
  #              validation type (REQUIRED, OPTIONAL, FORBIDDEN),
  #              each a list of the extensions of that type.
  #
  # Inputs:
  #   data_id (lower case)
  #
  # Outputs:
  #   ext_validation - a list of 3 validation type lists (required, optional,
  #                                                       forbidden)
  #
  # Exceptions:
  #   UnsupportedExtValidationType
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 01/07/04 49551  MSwam     first version
  #
  #-------------------------------------------------------------------------
    """
    # def find_ext_validation(self, data_id):
    #
    # result is a list of 3 empty lists to start
    # (REQUIRED, OPTIONAL and FORBIDDEN lists)
    ext_validation = [[],[],[]]
    for v in self._MissionObj._valid_extensions:
      if v['vae_data_id'] == data_id:
        #
        # MATCH found: add the extension to the set for that type
        #
        opusutil.PrintMsg("D","saving validation test: "+str(v))
        if v['vae_valid_type'] == "R":
          # REQUIRED, add to list
          ext_validation[REQUIRED].append(v['vae_extension'])
        elif v['vae_valid_type'] == "O":
          # OPTIONAL, add to list
          ext_validation[OPTIONAL].append(v['vae_extension'])
        elif v['vae_valid_type'] == "F":
          # FORBIDDEN, add to list
          ext_validation[FORBIDDEN].append(v['vae_extension'])
        else:
          # UNKNOWN
          opusutil.PrintMsg("E","Unsupported extension validation type: "+
                            str(v['vae_valid_type']))
          raise UnsupportedExtValidationType
    #
    return ext_validation


  def extValidationNeeded(self, datasetObj):
    """
  #-------------------------------------------------------------------------
  # 
  # Name: extValidationNeeded
  #
  # Description: Tests whether the given archive class is supposed to
  #              skip file extension validation, through its presence
  #              in a comma-separated list of classes to skip.
  #
  # Inputs:
  #   datasetObj
  #
  # Returns:
  #   1 = extension validation IS needed for this class
  #   0 = extension validation should be SKIPPED for this class
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 06/26/06 56017  MSwam     first version
  #
  #-------------------------------------------------------------------------
    """
    # def extValidationNeeded(self, datasetObj):
    #
    #
    # Some archive classes do not get their extensions validated (they
    # are typically part of a MULTICLASS and get validated in 
    # another class within the set)
    #
    if (self._classesSkipExtValidation.find(datasetObj._archive_class_name) 
        != -1):
      opusutil.PrintMsg("D","No extension validation needed for class "+
                        datasetObj._archive_class_name)
      for f in datasetObj._Files:
        f._extension_type = "O"
      return 0
    #
    return 1


  def getExtValidation(self, data_id):
    """
  #-------------------------------------------------------------------------
  # 
  # Name: getExtValidation
  #
  # Description: Locates and confirms presence of file extension validation
  #              info for a data id.
  #
  # Inputs:
  #   data_id
  #
  # Outputs:
  #   ext_validation
  #
  # Exceptions:
  #   MissingExtensionValidation
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 06/26/06 56017  MSwam     first version
  #
  #-------------------------------------------------------------------------
    """
    # def getExtValidation(self, data_id):
    #
    #
    opusutil.PrintMsg("D","find_ext_validation for "+data_id)
    ext_validation = self.find_ext_validation(data_id)

    if (len(ext_validation[REQUIRED]) == 0 and
        len(ext_validation[OPTIONAL]) == 0 and
        len(ext_validation[FORBIDDEN]) == 0):

      opusutil.PrintMsg("E","Unable to find extension validation info for "+
                             data_id)
      raise MissingExtensionValidation
    #
    return ext_validation


  def allExtAccepted(self, ext_validation, data_id, datasetObj):
    """
  #-------------------------------------------------------------------------
  # 
  # Name: allExtAccepted
  #
  # Description: Determines whether ALL file extensions are considered valid
  #              for a data ID, which is specified by a single, wild-carded
  #              required extension name.
  #
  # Inputs:
  #   ext_validation
  #   data_id
  #   datasetObj
  #
  # Outputs:
  #   1 = all extensions accepted (no extension validation required)
  #   0 = extension validation still required
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 06/26/06 56017  MSwam     first version
  #
  #-------------------------------------------------------------------------
    """
    # def allExtAccepted(self, ext_validation, data_id, datasetObj):
    #
    # if there is a REQUIRED validation type that is wildcarded then
    # all extensions are valid (considered OPTIONAL) and there is 
    # nothing to check in the files
    #
    if (len(ext_validation[REQUIRED]) == 1 and 
        ext_validation[REQUIRED][0] == "*"):

      opusutil.PrintMsg("D",data_id+": all extensions accepted, "
                            "no filtering needed")

      for f in datasetObj._Files:
        f._extension_type = "O"
      return 1
    #
    return 0


  def assignValidationCode(self, f, ext_validation, data_id):
    """
  #-------------------------------------------------------------------------
  # 
  # Name: assignValidationCode
  #
  # Description: Sets the file extension type and validity flag for a file,
  #              depending on whether the file's extension is required,
  #              optional, or forbidden.
  #
  # Inputs:
  #   f - FileList object
  #   ext_validation - lists of extension types to validate
  #   data_id - type of data being ingested
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 06/26/06 56017  MSwam     first version
  #
  #-------------------------------------------------------------------------
    """
    # def assignValidationCode(self, f, ext_validation, data_id):
    #
    if f._file_extension in ext_validation[REQUIRED]:
      f._extension_type = "R"
      opusutil.PrintMsg("D",data_id+": REQUIRED extension: "+f._file_extension)

    elif f._file_extension in ext_validation[OPTIONAL]:
      f._extension_type = "O"
      opusutil.PrintMsg("I",data_id+": OPTIONAL extension: "+f._file_extension)

    elif (f._file_extension in ext_validation[FORBIDDEN] or
          len(ext_validation[OPTIONAL]) == 0):
      # if there are no optional extensions, anything not required
      # is forbidden and any forbidden extension found is invalid
      f._extension_type = "F"
      f._valid_extension = 0
      opusutil.PrintMsg("E",data_id+": FORBIDDEN extension: "+f._file_extension)

    elif (len(ext_validation[OPTIONAL]) == 1 and
          ext_validation[OPTIONAL][0] == "*"):
      # anything neither REQUIRED nor FORBIDDEN defaults to OPTIONAL 
      # when the OPTIONAL type is wildcarded
      opusutil.PrintMsg("I",data_id+": OPTIONAL extension: "+f._file_extension)
      f._extension_type = "O"

    else:
      # an unexpected file extension was found
      f._extension_type = "U"
      f._valid_extension = 0
      opusutil.PrintMsg("E",data_id+": UNEXPECTED extension? "+f._file_extension)
    #
    return


  def markMissingRequiredExt(self, ext_validation, data_id, datasetObj):
    """
  #-------------------------------------------------------------------------
  # 
  # Name: markMissingRequiredExt
  #
  # Description: Marks a dataset as missing an extension if any of the
  #              required extensions are not present.
  #
  # Inputs:
  #   ext_validation
  #   data_id
  #   datasetObj
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 06/26/06 56017  MSwam     first version
  #
  #-------------------------------------------------------------------------
    """
    # def markMissingRequiredExt(self, ext_validation, data_id, datasetObj):
    #
    # if there are any REQUIRED extensions, they all MUST be present
    for ext in ext_validation[REQUIRED]:
      found = 0
      for f in datasetObj._Files:
        if ext == f._file_extension:
          found = 1
          break
      if not found:
        opusutil.PrintMsg("E",data_id+": Missing REQUIRED extension: "+ext)
        datasetObj._missing_ext = 1
    #
    return


  def validate_file_extensions(self, data_id):
    """
  #-------------------------------------------------------------------------
  # 
  # Name: validate_file_extensions
  #
  # Description: Validates the file extensions of all files in a
  #              dataset by using the valid_extensions information
  #              read from the db at initialization.
  #
  # Inputs:
  #   data_id
  #
  # Exceptions:
  #   ExtensionValidationFailed
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 01/07/04 49551  MSwam     first version
  # 06/19/06 56017  MSwam     pull unvalidated classes into class parameter
  #
  #-------------------------------------------------------------------------
    """
    # def validate_file_extensions(self, data_id):
    #
    # loop through the datasets
    for datasetObj in self._Datasets:
      #
      if not self.extValidationNeeded(datasetObj):
        continue
      
      opusutil.PrintMsg("I","validating extensions for "+
                            datasetObj._dataset_name+
                            ":"+datasetObj._archive_class_name)
  
      ext_validation = self.getExtValidation(data_id)

      opusutil.PrintMsg("D","found ext_validation for "+
                            data_id+" = "+str(ext_validation))
      
      if self.allExtAccepted(ext_validation, data_id, datasetObj):
        continue

      # categorize each file extension using the validation criteria
      #
      for f in datasetObj._Files:
        self.assignValidationCode(f, ext_validation, data_id)

      self.markMissingRequiredExt(ext_validation, data_id, datasetObj)

    #
    # loop through the datasets and files again, and raise an error if
    # any validation failures occurred
    #
    for datasetObj in self._Datasets:
      if datasetObj._missing_ext:
        raise ExtensionValidationFailed
      for f in datasetObj._Files:
        if not f._valid_extension:
          raise ExtensionValidationFailed
    #
    return


  def clearLists(self):
    """
  #-------------------------------------------------------------------------
  # 
  # Name: clearLists
  #
  # Description: Deletes Dataset and FileList objects.  Can be used
  #              to ensure each new event starts with clean lists.
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 06/26/06 56017  MSwam     first version
  #
  #-------------------------------------------------------------------------
    """
    # def clearLists(self):
    #
    # clear dataset and file lists
    for d in self._Datasets:
      del d._Files
    del self._Datasets
    self._Datasets = []
    #
    return


  def inQueryLogMode(self, dataset, data_id):
    """
  #-------------------------------------------------------------------------
  # 
  # Name: inQueryLogMode
  #
  # Description: If the requested data_id is in the list of install IDs,
  #              then we are determined to be in QUERY_LOG mode and the
  #              query log is written to for this dataset.
  # Inputs:
  #   dataset
  #   data_id
  #
  # Outputs:
  #   modeFlag = indicates QUERY_LOG or GROUP mode
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 06/26/06 56017  MSwam     first version
  #
  #-------------------------------------------------------------------------
    """
    # def inQueryLogMode(self, dataset, data_id):
    #
    # detect processing mode by checking the data_id against the install_ids
    if self._install_ids.find(data_id) != -1:
      # a match was found, so this is query and log mode
      # This mode only retrieves ingest_data_set_info fields and writes them
      # to the log and exits.  No file checking is performed in this mode.
      #
      opusutil.PrintMsg("I","QUERY/LOG mode.")
      self.query_log(dataset)
      return QUERY_LOG_MODE
    #
    return GROUP_MODE
    

  def setRequestIDs(self):
    """
  #-------------------------------------------------------------------------
  # 
  # Name: setRequestIDs
  #
  # Description: Sets the install request ID from the current time and
  #              a sequence number for each dataset.  Creates a sub-directory
  #              in the install area for this date.
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 06/26/06 56017  MSwam     first version
  #
  #-------------------------------------------------------------------------
    """
    # def setRequestIDs(self):
    #
    # set install request id for each dataset and save daystamp of last
    # (use same time for all datasets)
    time_now = time.time()
    for d in self._Datasets:
      d._dads_request_id = self.calcInstallRequestID(
                                                    time_now, d._sequence)
      opusutil.PrintMsg("D","install_request_id="+str(d._dads_request_id))
      daystamp = d._dads_request_id[5:13]

      # create ingest subdirectory if needed 
      #   (one per day for each archive class)
      #
      d._install_dir = (self._ingest_path + os.sep + 
                        d._archive_class_name.lower() + 
                        os.sep + daystamp + os.sep)
      opusutil.PrintMsg("D","subdir for ingest="+d._install_dir)
      if not os.path.isdir(d._install_dir):
        os.mkdir(d._install_dir)
    #
    return


  def moveFilesForIngest(self):
    """
  #-------------------------------------------------------------------------
  # 
  # Name: moveFilesForIngest
  #
  # Description: Moves each file in each dataset to the install directory.
  #
  # Exceptions:
  #   NoFilesFound
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 06/26/06 56017  MSwam     first version
  #
  #-------------------------------------------------------------------------
    """
    # def moveFilesForIngest(self):
    #
    # all checks have passed, move files into ingest subdirectory
    for d in self._Datasets:
      opusutil.PrintMsg("I","rename files for dataset "+str(d._dataset_name))
      for f in d._Files:
        fromfile = d._directory_path_name + os.sep + f._file_name
        tofile   = d._install_dir + os.sep + f._file_name
        opusutil.PrintMsg("I","rename from "+fromfile+" to "+tofile)
        try:
          os.rename(fromfile, tofile)
        except OSError, (errno, errstr):
          opusutil.PrintMsg("E","rename failed: "+errstr)
          raise NoFilesFound
    #
    return


  def computeGenDate(self):
    """ 
  #-------------------------------------------------------------------------
  # 
  # Name: computeGenDate
  #
  # Description: Determines and formats the generation date for this ingest
  #              request.
  #
  # Outputs: 
  #   generation_date
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 06/26/06 56017  MSwam     first version
  #
  #-------------------------------------------------------------------------
    """ 
    # def computeGenDate(self):
    #
    # determine request generation date across all files and
    # convert the numeric generation date into a readable date for the db
    numeric_gen_date = self.calcGenerationDate()
    generation_date = time.strftime("%b %d %Y %H:%M:%S", 
                                    time.localtime(numeric_gen_date))
    opusutil.PrintMsg("I","generation_date="+str(generation_date))
    #
    return generation_date


  def do_req_manager(self, dataset, data_id):
    """
  #-------------------------------------------------------------------------
  # 
  # Name: do_req_manager
  #
  # Description: Starts an ingest request, given the dataset and data_id.
  #
  # Input:
  #   dataset
  #   data_id
  #
  # Returns
  #   1 = success in Query/Log mode
  #   0 = success in Group mode
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 01/07/04 49551  MSwam     first version
  # 04/16/04 50823  J.Baum    remove time_now arg from 
  #                             load_ingest_data_set_info 
  # 12/03/07 58974  MSwam     move ingest_data_set_info method to MissionObj
  # 01/04/10 67158  heller    make faster
  #
  #-------------------------------------------------------------------------
    """
    # def do_req_manager(self, dataset, data_id):
    #
    self.clearLists()

    if self.inQueryLogMode(dataset, data_id):
      return QUERY_LOG_MODE

    # This is the complete, file-checking mode.
    opusutil.PrintMsg("I","GROUP mode.")

    self._Datasets = self._MissionObj.fill_Files(
                                        self._rm_control, 
                                        self._file_name_formats,
                                        dataset, data_id, 
                                        self._ingest_path, self._pipe_id,
                                        self._product_long_ext,
                                        self._exposure_long_ext)

    self.validate_file_extensions(data_id)

    generation_date = self.computeGenDate()

    self.checkForDuplicates(dataset, data_id, generation_date)

    self.setRequestIDs()
    self.moveFilesForIngest()

    # create db records
    self._MissionObj.load_ingest_data_set_info(
                                 dataset, data_id, 
                                 generation_date, self._Datasets, self._mission,
                                 self._rm_control[data_id]['irc_data_source'])
    self.load_ingest_files(generation_date)

    # just create the OSF with the final status to save time
    self.createOSFs(data_id)
    #self.updateOSFs(data_id)

    return GROUP_MODE


  def determineRetryMode(self, dataset, data_id):
    """ 
  #-------------------------------------------------------------------------
  # 
  # Name: determineRetryMode
  #
  # Description: Determines whether this retry request is for GROUP or
  #              INSTALL mode by checking the data_id against the list of
  #              INSTALL ids. In INSTALL mode the group info is looked up
  #              in the IDS (ingest_data_set_info) table.
  #
  # Input:
  #   dataset
  #   data_id
  #
  # Returns
  #   retry_mode
  #   dataset
  #   data_id
  #
  # Exceptions:
  #   MissingInfoRecord - if the IDS record for a request cannot be found
  #                       in the DB
  #   
  # History:
  # -------- -----  --------  --------------------------------------
  # 06/26/06 56017  MSwam     first version
  #
  #-------------------------------------------------------------------------
    """ 
    # def determineRetryMode(self, dataset, data_id):
    #
    retry_mode = RETRY_GROUP_OSF   # assume, until check for INSTALL is made
    #
    # detect processing mode by checking the data_id against the install_ids
    if self._install_ids.find(data_id) != -1:
      #
      # a match was found, so the retrigger was an INSTALL OSF
      # This needs to be "coverted" into the corresponding group OSF
      opusutil.PrintMsg("I","RETRY requested for INSTALL OSF: "+
                        dataset+","+data_id)
      opusutil.PrintMsg("I","Looking for corresponding GROUP info...")
      #
      # set a flag so that the GROUP osf is updated for "normal"
      # processing at the end of the retry processing (this happens
      # regularly back in the main event loop for a GROUP OSF RETRY, but
      # will NOT happen for an INSTALL OSF RETRY, since the normal
      # event handling in that case is for the INSTALL OSF, not the GROUP OSF)
      retry_mode = RETRY_INSTALL_OSF
      #
      # look up INSTALL id in the database to get group name and id
      ids_values = self._arch_dbobj.get_ingest_data_set_info_reqid(dataset)
      #
      # if not found, print error and raise exception
      if len(ids_values) == 0:
        opusutil.PrintMsg("E","Failed to find ingest_data_set_info record for "+
                          dataset)
        opusutil.PrintMsg("E","RETRY cannot be performed for this case.")
        raise MissingInfoRecord
      #
      # save and print the values
      dataset = str(ids_values[0][0])
      data_id = str(ids_values[0][1])
      opusutil.PrintMsg("I","Corresponding group info: "+ dataset +","+data_id)
    else:
      # no match of data_id in install_ids, so this is group mode retry
      opusutil.PrintMsg("I","RETRY GROUP mode.")
    #
    return (retry_mode, dataset, data_id)


  def getGenDate(self, dataset, data_id):
    """
  #-------------------------------------------------------------------------
  # 
  # Name: getGenDate
  #
  # Description: Retrieves generation date for an ingest request from
  #              the IDS DB table.
  #
  # Input:
  #   dataset
  #   data_id
  #
  # Returns
  #   gen_date
  #
  # Exceptions:
  #   MissingInfoRecord - no DB row found in ingest_data_set_info for
  #                       this dataset
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 06/26/06 56017  MSwam     first version
  #
  #-------------------------------------------------------------------------
    """
    # def getGenDate(self, dataset, data_id):
    #
    # get gen-date of last ingest request for this data
    ids_gen_date = self._arch_dbobj.ids_latest_receipt(self._mission, dataset, 
                                                 data_id)
    if len(ids_gen_date) == 0:
      opusutil.PrintMsg("E","Failed to find latest receipt "+
                            "ingest_data_set_info record for "+
                            dataset+","+data_id)
      opusutil.PrintMsg("E","RETRY cannot be performed for this case.")
      raise MissingInfoRecord
    gen_date = ids_gen_date[0][0]
    opusutil.PrintMsg("D","gen_date = "+gen_date)
    #
    return gen_date


  def getIDSfields(self, dataset, data_id, gen_date):
    """
  #-------------------------------------------------------------------------
  # 
  # Name: getIDSfields
  #
  # Description: Retrieves and verifies the ingest_data_set_info fields
  #              from the DB for a dataset/data_id/gen_date combination.
  #
  # Input:
  #   dataset
  #   data_id
  #   gen_date
  #
  # Exceptions:
  #   MissingInfoRecord
  #   MismatchedRecordCount
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 06/26/06 56017  MSwam     first version
  # 10/05/06 56017  MSwam     fix bug
  #
  #-------------------------------------------------------------------------
    """
    # def getIDSfields(self, dataset, data_id, gen_date):
    #
    # get all records for this dataset and gen_date
    ids_fields = self._arch_dbobj.get_ids_fields_gen_date(self._mission, 
                                                          dataset, data_id, 
                                                          gen_date)
    if len(ids_fields) == 0:
      opusutil.PrintMsg("E","Failed to find "+
                            "ingest_data_set_info record for "+
                            dataset+","+data_id+","+gen_date)
      raise MissingInfoRecord
    opusutil.PrintMsg("D","ids_fields= "+str(ids_fields))

    # verify record counts match between ids and request_members
    if len(ids_fields) != len(self._Datasets):
      opusutil.PrintMsg("E","MISMATCH in record count between "+
                            "ingest_data_set_info and request_members for "+
                            dataset+","+data_id+","+gen_date)
      opusutil.PrintMsg("E","len,ids_fields= "+str(len(ids_fields))+","+
                            str(ids_fields))
      for d in self._Datasets:
        d.dump()
      raise MismatchedRecordCount
    #
    return ids_fields


  def getIFIrows(self, aSet, gen_date):
    """
  #-------------------------------------------------------------------------
  # 
  # Name: getIFIrows
  #
  # Description: Retrieves the ingest_file_info (IFI) rows from the DB
  #              for a dataset/ARCHIVE_CLASS/generation date combination.
  #
  # Input:
  #   aSet - a Datasets object
  #   gen_date - the generation date of the dataset
  #
  # Returns
  #   theRows - the IFI rows from the DB
  #
  # Exceptions:
  #   MissingInfoRecord - no IFI rows found for the request
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 06/26/06 56017  MSwam     first version
  #
  #-------------------------------------------------------------------------
    """
    # def getIFIrows(self, aSet, gen_date):
    #
    # find the files in ingest_files table
    theRows = self._arch_dbobj.get_ifi_rows(self._mission, 
                                            aSet[IDS_DATASETNAME].upper(),
                                            aSet[IDS_ARCHIVECLASS],
                                            gen_date)
    if len(theRows) == 0:
      opusutil.PrintMsg("E","Failed to find ingest_files record for "+
                         aSet[IDS_DATASETNAME]+","+
                         aSet[IDS_ARCHIVECLASS]+","+
                         gen_date)
      raise MissingInfoRecord
    opusutil.PrintMsg("I","ifi rows="+str(theRows))
    #
    return (theRows)


  def renameIngestFiles(self, aSet, theRows):
    """
  #-------------------------------------------------------------------------
  # 
  # Name: renameIngestFiles
  #
  # Description: Renames each file saved in a set of ingest_file_info
  #              rows from the date-stamped ingest subdirectory UP one
  #              level (preparing it for a re-ingest attempt).  The last
  #              file moved has its mod date reset so that a new generation
  #              date will be calculated for this re-ingest attempt.
  #
  # Input:
  #   aSet
  #   theRows
  #
  # Exceptions:
  #   NoFilesFound
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 06/26/06 56017  MSwam     first version
  #
  #-------------------------------------------------------------------------
    """
    # def renameIngestFiles(self, aSet, theRows):
    #
    for aRow in theRows:
      fromfile = aSet[IDS_PATHNAME]+str(aRow[0])
      tofile = aSet[IDS_PATHNAME][:-9]+str(aRow[0]) # trimmed YYYYMMDD/
      opusutil.PrintMsg("I","renaming "+fromfile+ " to "+tofile)
      try:
        os.rename(fromfile, tofile)
      except OSError, (errno, errstr):
        opusutil.PrintMsg("E","rename failed: "+errstr)
        raise NoFilesFound
    #
    # "touch" the last file moved to reset its mod date
    os.utime(tofile, None)
    #
    return


  def updateOSFtoClean(self, aSet):
    """
  #-------------------------------------------------------------------------
  # 
  # Name: updateOSFtoClean
  #
  # Description: Updates an OSF for a dataset to trigger the clean-up stage.
  #
  # Input:
  #   aSet
  #
  # Exceptions:
  #   OSFUpdateFailed
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 06/26/06 56017  MSwam     first version
  #
  #-------------------------------------------------------------------------
    """
    # def updateOSFtoClean(self, aSet):
    #
    #
    # update the original install OSF for cleaning
    try:
      opusutil.PrintMsg("I","updating OSF for dataset="+
                            aSet[IDS_REQUESTID]+
                            " path="+self._pathname)
      os.system('osf_update -p ' + self._pathname + 
                ' -f ' + aSet[IDS_REQUESTID]+
                ' -c ' + self._cleanup_osf_stage + 
                ' -s ' + self._cleanup_osf_status)
    except:
      opusutil.PrintMsg('E','OSF update FAILED for '+
                            aSet[IDS_REQUESTID])
      raise OSFUpdateFailed
    #
    return


  def updateGroupOSF(self, dataset, data_id):
    """
  #-------------------------------------------------------------------------
  # 
  # Name: updateGroupOSF
  #
  # Description: Updates the OSF for a dataset group to causing a retry
  #              of ingest.
  #
  # Input:
  #   dataset
  #   data_id
  #
  # Exceptions:
  #   OSFUpdateFailed
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 06/26/06 56017  MSwam     first version
  #
  #-------------------------------------------------------------------------
    """
    # def updateGroupOSF(self, dataset, data_id):
    #
    # update the group osf to cause "normal" retriggering using the
    # OSF_UPDATE resource value
    try:
      opusutil.PrintMsg("I","updating OSF for dataset="+dataset+","+data_id+
                             " path="+self._pathname)
      os.system('osf_update -p ' + self._pathname + 
                 ' -f ' + dataset +
                 ' -t ' + data_id +
                 ' -s ' + self._osf_update_status)
    except:
      opusutil.PrintMsg('E','OSF update FAILED for '+dataset+','+data_id)
      raise OSFUpdateFailed
    #
    return


  def retry(self, dataset, data_id):
    """
  #----------------------------------------------------------------------
  #
  # Name: retry
  #
  # Description: Sets up for retry of an ingest request, given the dataset and 
  #              data_id.  This results in
  #			- determining the processing mode and finding the
  #				group OSF
  #			- determining the request type, dataset names, and
  #				archive class
  #			- moving all dataset files out of dated subdir
  #			- resetting file mod time for one dataset
  #			- triggering CLNING by updating install OSFs
  #              A number of verification checks are performed during the
  #              process to make sure that the ingest request being retried
  #              matches up with the database contents.
  # Input:
  #   dataset
  #   data_id
  #
  # Returns: 
  #   retry_mode, set to RETRY_GROUP_OSF or RETRY_INSTALL_OSF
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 06/28/04 50703  MSwam     add retry mode
  # 10/05/06 56017  MSwam     fix bug
  #
  #-------------------------------------------------------------------------
    """
    # def retry(self, dataset, data_id):
    #
    self.clearLists()

    (retry_mode, dataset, data_id) = self.determineRetryMode(dataset, data_id)

    self._Datasets = self._MissionObj.fill_Datasets(self._rm_control,
                                                    dataset, data_id,
                                                    self._ingest_path,
                                                    self._pipe_id)

    gen_date = self.getGenDate(dataset, data_id)

    ids_fields = self.getIDSfields(dataset, data_id, gen_date)

    # for each dataset, find the files and move them to the class-specific
    # archive directory from their time-stamped subdir
    #
    for aSet in ids_fields:
       opusutil.PrintMsg("I","aSet="+str(aSet))

       theRows = self.getIFIrows(aSet, gen_date)

       self.renameIngestFiles(aSet, theRows)

       self.updateOSFtoClean(aSet)
    #
    self._arch_dbobj.update_ids_for_cleaning(self._mission, dataset, data_id,
                                             gen_date)
    if retry_mode == RETRY_INSTALL_OSF:
      self.updateGroupOSF(dataset, data_id)
    #
    return retry_mode


  def processHaltEvent(self, event):
    """
  #-------------------------------------------------------------------------
  # Name: processHaltEvent
  #
  # Description: Handle a HALT event (override base class).
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 07/02/04 50703  MSwam     first version
  #
  #-------------------------------------------------------------------------
    """
    # def processHaltEvent(self, event):
    #
    opusutil.PrintMsg("I","Received HALT event. Stopping.")
    self._arch_dbobj.close()
    raise HaltEvent


  def closeOSF(self, eventTag, event):
    """
  #-------------------------------------------------------------------------
  #
  # Name: closeOSF
  #
  # Description: Common code for closing an event and detecting a failure.
  #
  # Inputs:
  #   eventTag - OSF state to close the event with
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 06/26/06 56017  MSwam     first version
  #
  #-------------------------------------------------------------------------
    """
    # def closeOSF(self, eventTag, event):
    #
    traceback.print_exc()
    try:
      opusutil.PrintMsg("E","closing event with "+eventTag)
      self._papp.closeEvent(eventTag, event.eid)
    except:
      traceback.print_exc()
      opusutil.PrintMsg("E","Failed to close event "+eventTag)
    #
    return


  def finishTrl(self, processName, dataset, data_id):
    """
  #-------------------------------------------------------------------------
  #
  # Name: finishTrl
  #
  # Description: Common code to write the final processing message for
  #              a dataset and close its log file.
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 06/26/06 56017  MSwam     first version
  #
  #-------------------------------------------------------------------------
    """
    # def finishTrl(self, processName, dataset, data_id):
    opusutil.PrintMsg("I","----- "+processName+" ended for: "+dataset+"_"+
                                                              data_id+" -----")
    opusutil.CloseTrl()
    return


  def processOsfEvent(self, event):
    """
  #-------------------------------------------------------------------------
  # Name: processOsfEvent
  #
  # Description: Handle an OSF event (override base class)
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 07/02/04 50703  MSwam     first version
  # 07/22/04 50703  MSwam     rework INSTALL retry event handling
  #
  #-------------------------------------------------------------------------
    """
    #
    # def processOsfEvent(self, event):
    dataset, data_id = self.getDatasetDataidFromOsfEvent(event)

    try:
      opusutil.OpenTrl(self._ingest_log_dir + os.sep + dataset + "_" +
                       data_id + ".log")
    except:
      # any exception marks the OSF with OSF_FAIL
      self.closeOSF("OSF_FAIL", event)
      return

    opusutil.PrintMsg("I","----- RMGR started for: "+dataset+"_"+ 
                                                     data_id+" -----")
    retry_mode = RETRY_NONE # default
    try:
      if self._eventTriggerName == "OSF_TRIGGER2":  # RETRY trigger
        retry_mode = self.retry(dataset, data_id)
        #
        # no exception means we are ready to retry this request,
        # IF this was a GROUP retry. 
        #
        # INSTALL retries should NOT
        # finish the event loop normally, since the old event no longer 
        # applies, and might even have been cleaned away by CLNING,
        # so they get a "dead end" status.
        #
        if retry_mode == RETRY_GROUP_OSF:
          success_event = "OSF_TRIGGER1"

        else: # RETRY_INSTALL_OSF
          success_event = "OSF_DEAD_END"
      #
      else:                                       # OSF_TRIGGER1 = normal mode
        mode = self.do_req_manager(dataset, data_id)
        if mode == QUERY_LOG_MODE:
          success_event = "OSF_SUCCESS"
        else:
          # Group mode
          success_event = "OSF_SKIP"

    except DuplicateRequest:
      #
      self.closeOSF("OSF_DUPLICATE", event)
      self.finishTrl("RMGR", dataset, data_id)
      return

    except: # any other exception treated as failure
      self.closeOSF("OSF_FAIL", event)
      self.finishTrl("RMGR", dataset, data_id)
      return

    try:
      self._papp.closeEvent(success_event, event.eid)
    except:
      if retry_mode != RETRY_INSTALL_OSF:
        #
        # The RETRY_INSTALL_OSF mode will usually generate an exception
        # because the INSTALL OSF that was retriggered gets deleted in
        # the cleanup process (if CLNING is running), and is no longer 
        # around to have its event properly closed.  The messages will be 
        # suppressed for that case.
        #
        traceback.print_exc()
        opusutil.PrintMsg("E","Failed to close "+success_event+" event")
    #
    self.finishTrl("RMGR", dataset, data_id)
    return


#-------------------------------------------------------------------------
# The main routine. Handles interactive and pipeline mode.
#
# History:
# -------- -----  --------  --------------------------------------
# 01/07/04 49551  MSwam     first version
# 07/02/04 50703  MSwam     use new base_poller base class
#
#-------------------------------------------------------------------------
#
# initialize
try:
  poller = ReqManager()
  poller.initialize()
  poller.pipelineInit()
#
except base_poller.InteractiveMode:
  opusutil.PrintMsg("I","Running command-line mode.")
  try:
    poller.interactive()
  except UsageError:
    sys.exit(NOT_OK)
  except:
    traceback.print_exc()
    opusutil.PrintMsg("F","Command-line mode Failed.")
    sys.exit(NOT_OK)
  sys.exit(OK)
#
except:
  traceback.print_exc()
  opusutil.PrintMsg("E","initialization failed")
  sys.exit(NOT_OK)

opusutil.PrintMsg("I","Pipeline set-up complete.")
try:
  poller.run()
except HaltEvent:
  sys.exit(OK)
except:
  traceback.print_exc()
  opusutil.PrintMsg("E","pipeline_process failed")
  sys.exit(NOT_OK)
